/********************************************************\
|*                                                      *|
|*                  VeriFinger 5.0 SDK                  *|
|*                                                      *|
|*  ScanMan.h                                           *|
|*    Header file for Scanner Manager library           *|
|*                                                      *|
|*  Copyright (C) 2002-2006 Neurotechnologija           *|
|*                                                      *|
\********************************************************/


// Avoid multiple includes
#ifndef __SCANMAN_INCLUDED__
#define __SCANMAN_INCLUDED__


// Force C (not C++) names for Scanner Manager API functions
#ifdef __cplusplus
extern "C"
{
#endif


///////////
//       //
// Types //
//       //
///////////

typedef unsigned char BYTE;
typedef signed char SBYTE;
typedef char CHAR;
typedef unsigned short WORD;
typedef unsigned short USHORT;
typedef short SHORT;
typedef unsigned long DWORD;
typedef unsigned int UINT;
typedef unsigned long ULONG;
typedef int INT;
typedef long LONG;
typedef float FLOAT;
typedef double DOUBLE;
typedef int BOOL;


#ifndef FALSE
#define FALSE 0
#endif

#ifndef TRUE
#define TRUE 1
#endif

#ifndef VOID
#define VOID void
#endif

#ifndef NULL
#ifdef __cplusplus
#define NULL 0
#else
#define NULL ((void*)0)
#endif
#endif


////////////
//        //
// Consts //
//        //
////////////

// Parameter types
#define SM_TYPE_VOID	0
#define SM_TYPE_BYTE	1
#define SM_TYPE_SBYTE	2
#define SM_TYPE_WORD	3
#define SM_TYPE_SHORT	4
#define SM_TYPE_DWORD	5
#define SM_TYPE_INT		6
#define SM_TYPE_BOOL	10
#define SM_TYPE_CHAR	20
#define SM_TYPE_STRING	100

// Events
#define SM_EVENT_PLUGGED	1
#define SM_EVENT_UNPLUGGED	2
#define SM_EVENT_ERROR		3

// States
#define SM_STATE_FINGER_DOWN	1
#define SM_STATE_FINGER_UP		2


///////////////
//           //
// Callbacks //
//           //
///////////////

#define SM_CALLBACK __stdcall

typedef VOID SM_CALLBACK SM_MONITOR_PROC(const CHAR * scanner_id, DWORD event, VOID * param);

typedef VOID SM_CALLBACK SM_IMAGE_PROC(const CHAR * scanner_id, INT width, INT height, const BYTE * image, INT resolution, VOID * param);
typedef VOID SM_CALLBACK SM_STATE_PROC(const CHAR * scanner_id, DWORD state, VOID * param);


///////////////
//           //
// Functions //
//           //
///////////////

#define SCANMAN_API __stdcall


// Initialization
INT SCANMAN_API SMInitialize();
INT SCANMAN_API SMFinalize();

// Parameters
INT SCANMAN_API SMGetParameter(INT reserved, INT parameter, VOID * value);
INT SCANMAN_API SMSetParameter(INT reserved, INT parameter, DWORD value);

// Scanner enumeration
INT SCANMAN_API SMGetScannerCount();
INT SCANMAN_API SMGetScannerId(INT index, CHAR *id, INT len);

// Scanner monitoring
INT SCANMAN_API SMSetMonitor(SM_MONITOR_PROC * monitor_proc, VOID * param);
INT SCANMAN_API SMRemoveMonitor();

// Capturing
INT SCANMAN_API SMStartCapturing(const CHAR * scanner_id, SM_IMAGE_PROC * image_proc, SM_STATE_PROC * state_proc, VOID * param);
INT SCANMAN_API SMStopCapturing(const CHAR * scanner_id);


/////////////////
//             //
// Error codes //
//             //
/////////////////

// Pass SMXxx() function result to this macro to determine if the execution of that function failed
#define SMFailed(result) ((result) < 0)

// Pass SMXxx() function result to this macro to determine if the execution of that function succeeded
#define SMSucceeded(result) ((result) >= 0)

// General
#define SME_OK					0
#define SME_FAILED				-1
#define SME_OUT_OF_MEMORY		-2
#define SME_NOT_INITIALIZED		-3
#define SME_ARGUMENT_NULL		-4
#define SME_INVALID_ARGUMENT	-5
// Parameters
#define SME_INVALID_PARAMETER	-10
#define SME_PARAMETER_READ_ONLY	-11
// Capturing
#define SME_NOT_CAPTURING		-100
#define SME_ALREADY_CAPTURING	-101


////////////////
//            //
// Parameters //
//            //
////////////////

/**************************************************************************************\

Parameters can be of the following types:
	byte (SM_TYPE_BYTE)
	signed byte (SM_TYPE_SBYTE)
	word (SM_TYPE_WORD)
	short integer (SM_TYPE_SHORT)
	double word (SM_TYPE_DWORD)
	integer (SM_TYPE_INT)
	boolean (SM_TYPE_BOOL)
	char (SM_TYPE_CHAR);
	string (SM_TYPE_STRING)
For SMGetParameters "VOID * value" have to be:
	BYTE* for byte
	SBYTE* for signed byte
	WORD* for word
	SHORT* for short integer
	DWORD* for double word
	INT* for integer
	BOOL* for boolean
	CHAR* for char
	CHAR* for string (must point to array of at least string length + 1 CHARs; if NULL string length is returned)
For SMSetParameters "DWORD value" have to be:
	(DWORD)(BYTE) for byte;
	(DWORD)(SBYTE) for signed byte
	(DWORD)(WORD) for word
	(DWORD)(SHORT) for short integer
	(DWORD) for double word
	(DWORD)(INT) for integer
	(DWORD)(BOOL) for boolean
	(DWORD)(CHAR) for char
	(DWORD)(CHAR*) for string
Some parameters are read only

\**************************************************************************************/

// General
#define SMP_TYPE			0 // Get | Pass parameter identifier ((VOID*)SMP_XXX) as value | Returns one of SM_TYPE_XXX constants
#define SMP_NAME			10 // Get | String
#define SMP_VERSION_HIGH	11 // Get | DWord
#define SMP_VERSION_LOW		12 // Get | DWord
#define SMP_COPYRIGHT		13 // Get | String


#ifdef __cplusplus
}
#endif


#endif // !__SCANMAN_INCLUDED__
