#pragma once

template<class T> class CCollectionBase
{
private:
	bool freeOnRemove;
	vector<T *> items;

protected:
	vector<T *> & get_Items() { return items; }
	__declspec(property(get = get_Items)) vector<T *> & Items;

	virtual void OnClear()
	{
	}

	virtual void OnClearComplete()
	{
	}

	virtual void OnInsert(int /*index*/, T * value)
	{
		OnValidate(value);
	}
	
	virtual void OnInsertComplete(int /*index*/, T * value)
	{
		OnAdded(value);
	}
	
	virtual void OnRemove(int /*index*/, T * /*value*/)
	{
	}
	
	virtual void OnRemoveComplete(int /*index*/, T * value)
	{
		OnRemoved(value);
	}

	virtual void OnValidate(T * value)
	{
		if(!value) throw exception("Collection item can not be NULL");
	}

	virtual void OnAdded(T * /*value*/)
	{
	}

	virtual void OnRemoved(T * value)
	{
		if(freeOnRemove) delete value;
	}

public:
	CCollectionBase(bool freeOnRemove = false)
		: freeOnRemove(freeOnRemove)
	{
	}

	~CCollectionBase()
	{
		Clear();
	}

	int get_Count()
	{
		return items.Length();
	}
	__declspec(property(get = get_Count)) int Count;

	int get_IsEmpty()
	{
		return items.IsEmpty();
	}
	__declspec(property(get = get_Count)) int IsEmpty;

	void Clear()
	{
		OnClear();
		if(!items.IsEmpty())
			for(T * * pItem = items.End() - 1; pItem >= items.Begin(); pItem--)
				OnRemoved(*pItem);
		OnClearComplete();
	}

	int Add(T * value)
	{
		int index = items.Length();
		Insert(index, value);
		return index;
	}

	void Insert(int index, T * value)
	{
		OnInsert(index, value);
		items.Insert(index, value);
		OnInsertComplete(index, value);
	}

	void RemoveAt(int index)
	{
		T * value = items[index];
		OnRemove(index, value);
		items.Remove(index);
		OnRemoveComplete(index, value);
	}
	
	void Remove(T * value)
	{
		int index = items.IndexOf(value);
		if(index == -1) throw exception("Collection does not contain the item");
		RemoveAt(index);
	}

	bool Contains(T * value)
	{
		return items.Contains(value);
	}

	int IndexOf(T * value)
	{
		return items.IndexOf(value);
	}
};

template<class T> class CCollection: public CCollectionBase<T>
{
public:
	CCollection(bool freeOnRemove = false)
		: CCollectionBase<T>(freeOnRemove)
	{
	}

	T * get_Item(int index)
	{
		vector<T *> & items = Items;
		return items[index];
	}
	_declspec(property(get = get_Item)) T * Item[];
};

template<class T> class CCollectionWithSet: public CCollectionBase<T>
{
protected:
	virtual void OnSet(int index, T * oldValue, T * newValue)
	{
		#pragma warning(disable: 4100)
		OnValidate(newValue);
	}
	
	virtual void OnSetComplete(int index, T * oldValue, T * newValue)
	{
		#pragma warning(disable: 4100)
		OnRemoved(oldValue);
		OnAdded(newValue);
	}

public:
	CCollectionWithSet(bool freeOnRemove = false)
		: CCollectionBase<T>(freeOnRemove)
	{
	}

	T * get_Item(int index)
	{
		vector<T *> & items = Items;
		return items[index];
	}
	void set_Item(int index, T * value)
	{
		T * oldValue = items[index];
		if(value == oldValue) return;
		OnSet(index, oldValue, value);
		vector<T *> & items = Items;
		items[index] = value;
		OnSetComplete(index, oldValue, value);
	}
	_declspec(property(get = get_Item, put = set_Item)) T * Item[];
};
