#include "System.h"
#include "Log.h"
#include "Utils.h"

CLog::CLog()
{
	InitializeCriticalSection(&cs);
}

CLog::~CLog()
{
	CloseFile();
	DeleteCriticalSection(&cs);
}

void CLog::OpenFile(const string &fileName, LogMode mode)
{
	CloseFile();
	EnterCriticalSection(&cs);
	FileOpenMode openMode;
	if(mode == lmAppend) openMode = foAppend;
	else if(mode == lmOverwrite) openMode = foTruncateAlways;
	else throw exception("Invalid log mode");
	fileStream.Open(fileName, openMode, faWrite, fsRead);
	LeaveCriticalSection(&cs);
}

void CLog::CloseFile()
{
	EnterCriticalSection(&cs);
	fileStream.Close();
	LeaveCriticalSection(&cs);
}

void CLog::Flush()
{
	EnterCriticalSection(&cs);
	fileStream.Flush();
	LeaveCriticalSection(&cs);
}

void CLog::WriteLine(const TCHAR *format, ...)
{
	va_list list;
	va_start(list, format);
	AddLine(FormatV(format, list));
	va_end(list);
}

void CLog::WriteLine(DWORD id, ...)
{
	va_list list;
	va_start(list, id);
	AddLine(FormatV(id, list));
	va_end(list);
}

void CLog::Write(const TCHAR *format, ...)
{
	va_list list;
	va_start(list, format);
	Add(FormatV(format, list));
	va_end(list);
}

void CLog::Write(DWORD id, ...)
{
	va_list list;
	va_start(list, id);
	Add(FormatV(id, list));
	va_end(list);
}

void CLog::NewLine()
{
	AddLine(emptyString);
}

void CLog::AddLine(const string &text)
{
	EnterCriticalSection(&cs);
	OnAddLine(text);
	lastLine.Clear();
	LeaveCriticalSection(&cs);
}

void CLog::Add(const string &text)
{
	EnterCriticalSection(&cs);
	OnAdd(text);
	lastLine += text;
	LeaveCriticalSection(&cs);
}

void CLog::Clear()
{
	EnterCriticalSection(&cs);
	OnClear();
	lastLine.Clear();
	LeaveCriticalSection(&cs);
}

string CLog::LastLine()
{
	EnterCriticalSection(&cs);
	string s = lastLine;
	s.Update();
	LeaveCriticalSection(&cs);
	return s;
}

void CLog::OnClear()
{
	if(fileStream.Opened())
	{
		fileStream.Truncate();
	}
}

void CLog::OnAddLine(const string &text)
{
	if(fileStream.Opened()) fileStream.WriteLine(text);
}

void CLog::OnAdd(const string &text)
{
	if(fileStream.Opened()) fileStream.WriteString(text);
}

void CLog::CheckStatesEmpty()
{
	if(states.IsEmpty()) throw exception("States are empty");
}

void CLog::AddState(const string &value)
{
	if(value.IsEmpty()) throw exception("Can not add an empty log state");
	states.Add(LogState(value));
	UpdateState();
}

void CLog::RemoveState()
{
	CheckStatesEmpty();
	states.RemoveLast();
	UpdateState();
}

void CLog::InitStateProgressS(const vector<int> &counts, const vector<int> &countsStarted, bool asIndex)
{
	CheckStatesEmpty();
	if(countsStarted.Length() != counts.Length()) throw exception("Log state counts started has invalid size");
	LogState &state = states.Last();
	try
	{
		state.asIndex = asIndex;
		state.count = 0;
		state.progresses.SetLength(counts.Length());
		state.progress = 0;
		state.progressP = 0.0f;
		state.times.SetLength(counts.Length());
		state.times.Fill(0.0);
		state.time = 0.0;
		state.timesEstimated.SetLength(counts.Length());
		state.timesEstimated.Fill(0.0);
		state.timeEstimated = 0.0;
		state.timesRemaining.SetLength(counts.Length());
		state.timesRemaining.Fill(0.0);
		state.timeRemaining = 0.0;
		for(const int * i = counts.Begin(), * j = countsStarted.Begin(); i < counts.End(); i++, j++)
		{
			if(*i < 0) throw exception("One of log state counts is negative");
			if(*j < 0) throw exception("One of log state counts started is negative");
			if(*j > *i) throw exception("One of log state counts started is greater than count");
			state.count += *i;
		}
		state.counts = counts;
		state.countsStarted = countsStarted;
		SetStateProgressRaw(state, countsStarted);
		UpdateState();
	}
	catch(...)
	{
		state.asIndex = false;
		state.counts.Clear();
		state.count = 0;
		state.countsStarted.Clear();
		state.progresses.Clear();
		state.times.Clear();
		state.timesEstimated.Clear();
		state.timesRemaining.Clear();
		throw;
	}
}

void CLog::SetStateProgressRaw(LogState &state, const vector<int> &progresses)
{
	if(progresses.Length() != state.progresses.Length())
		throw exception("Invalid progress count in log state");
	try
	{
		state.progress = 0;
		for(const int * i = progresses.Begin(), * cs = state.countsStarted.Begin(); i < progresses.End(); i++, cs++)
		{
			if(*i < 0) throw exception("One of log state progresses is negative");
			if(*i < *cs) throw exception("One of log state progresses is less that count started");
			state.progress += *i;
		}
		state.progressP = (float)PercentFZ(state.progress, state.count);
		state.progresses = progresses;
	}
	catch(...)
	{
		state.progresses.Fill(0);
		state.progress = 0;
		state.progressP = 0.0f;
		state.timesEstimated.Fill(0.0);
		state.timeEstimated = 0.0;
		state.timesRemaining.Fill(0.0);
		state.timeRemaining = 0.0;
		throw;
	}
}

void CLog::SetStateTimeRaw(LogState &state, const vector<double> &times)
{
	if(times.Length() != state.times.Length())
		throw exception("Invalid time count in log state");
	try
	{
		state.time = 0.0;
		for(const double * i = times.Begin(); i < times.End(); i++)
		{
			if(*i < 0) throw exception("One of log state times is negative");
			state.time += *i;
		}
		state.times = times;
	}
	catch(...)
	{
		state.time = 0;
		state.timesEstimated.Fill(0.0);
		state.timeEstimated = 0.0;
		state.timesRemaining.Fill(0.0);
		state.timeRemaining = 0.0;
	}
}

void CLog::SetStateProgress(const vector<int> &progresses)
{
	CheckStatesEmpty();
	LogState &state = states.Last();
	SetStateProgressRaw(state, progresses);
	UpdateStateTime();
}

void CLog::SetStateTime(const vector<double> &times)
{
	CheckStatesEmpty();
	LogState &state = states.Last();
	SetStateTimeRaw(state, times);
	UpdateStateTime();
}

void CLog::SetStateProgressAndTime(const vector<int> &progresses, const vector<double> &times)
{
	CheckStatesEmpty();
	LogState &state = states.Last();
	SetStateProgressRaw(state, progresses);
	SetStateTimeRaw(state, times);
	UpdateStateTime();
}

void CLog::UpdateStateTime()
{
	LogState &state = states.Last();
	state.timeEstimated = 0.0;
	state.timeRemaining = 0.0;
	double * t = state.times.Begin(), * e = state.timesEstimated.Begin(), * r = state.timesRemaining.Begin();
	for(int * c = state.counts.Begin(), * cs = state.countsStarted.Begin(), * p = state.progresses.Begin(); c < state.counts.End(); c++, cs++, p++, t++, e++, r++)
	{
		*e = *p == *cs ? 0.0 : (*t) * (*c - *cs) / (*p - *cs);
		state.timeEstimated += *e;
		*r = *e - *t;
		state.timeRemaining += *r;
	}
	UpdateState();
}

void CLog::UpdateState()
{
	OnUpdateState(states);
}

void CLog::OnUpdateState(const vector<LogState> &/*states*/)
{
}
