/********************************************************\
|*                                                      *|
|*                  VeriFinger 4.2 SDK                  *|
|*                                                      *|
|*  VFingerX.cpp                                        *|
|*    Source file for extended VeriFinger features      *|
|*                                                      *|
|*  Copyright (C) 1998-2004 Neurotechnologija           *|
|*                                                      *|
\********************************************************/

#include "System.h"
#include "Utils.h"
#include <math.h>
#include "VFingerX.h"

////////////////////////////////
//                            //
// Error and result functions //
//                            //
////////////////////////////////

string VFErrorToString(INT error)
{
	if(error >= 0) return emptyString;
	switch(error)
	{
		case VFE_FAILED: return TEXT("Failed");
		case VFE_OUT_OF_MEMORY: return TEXT("Out of memory");
		case VFE_NOT_INITIALIZED: return TEXT("Not initialized");
		case VFE_ARGUMENT_NULL: return TEXT("Argument null");
		case VFE_INVALID_ARGUMENT: return TEXT("Invalid argument");
		case VFE_NOT_IMPLEMENTED: return TEXT("Not implemented");
		case VFE_NOT_REGISTERED: return TEXT("Not registered");
		case VFE_INVALID_SERIAL_NUMBER: return TEXT("Invalid serial number");
		case VFE_INVALID_REGISTRATION_KEY: return TEXT("Invalid registration key");
		case VFE_SCANNER_DRIVER_ERROR: return TEXT("Scanner driver error");
		case VFE_REGISTRATION_NOT_NEEDED: return TEXT("Registration not needed");
		case VFE_NO_SCANNER: return TEXT("No scanner connected");
		case VFE_MORE_THAN_ONE_SCANNER: return TEXT("More than one scanner connected");
		case VFE_LM_CONNECTION_ERROR: return TEXT("Communication with License Manager error");
		case VFE_LM_NO_MORE_LICENSES: return TEXT("License Manager error: no more licenses");
		case VFE_INVALID_PARAMETER: return TEXT("Invalid parameter");
		case VFE_PARAMETER_READ_ONLY: return TEXT("Parameter is read only");
		case VFE_ILLEGAL_IMAGE_RESOLUTION: return TEXT("Illegal image resolution");
		case VFE_ILLEGAL_IMAGE_SIZE: return TEXT("Illegal image size");
		case VFE_LOW_QUALITY_IMAGE: return TEXT("Low quality image");
		case VFE_INVALID_MODE: return TEXT("Invalid mode");
		case VFE_INVALID_FEATURES_FORMAT: return TEXT("Invalid features format");
		default: return IntToString(error);
	}
}

string VFResultToString(INT result)
{
	if(VFFailed(result)) return VFErrorToString(result);
	else return result == VFE_OK ? STRING("OK") : Format(TEXT("OK (%d)"), result);
}

void VFRaiseError(INT error)
{
	throw EXCEPTION(Format(VF_ENGINE_NAME TEXT(" error:\n\tCode = %d\n\tDescription = %s"),
		error, (LPCTSTR)VFErrorToString(error)));
}

void VFCheckResult(INT result)
{
	if(VFFailed(result)) VFRaiseError(result);
}

//////////////////////////
//                      //
// Parameters functions //
//                      //
//////////////////////////

bool VFParameterOfType(INT parameter, INT type)
{
	return VFGetParameterType(parameter) == type;
}

void VFCheckParameterType(INT parameter, INT type)
{
	if(!VFParameterOfType(parameter, type)) throw exception("VeriFinger parameter is of wrong type");
}

// Get

BYTE VFGetByteParameter(INT parameter, HVFCONTEXT context)
{
	VFCheckParameterType(parameter, VF_TYPE_BYTE);
	BYTE value;
	VFCheckResult(VFGetParameter(parameter, &value, context));
	return value;
}

SBYTE VFGetSByteParameter(INT parameter, HVFCONTEXT context)
{
	VFCheckParameterType(parameter, VF_TYPE_SBYTE);
	SBYTE value;
	VFCheckResult(VFGetParameter(parameter, &value, context));
	return value;
}

WORD VFGetWordParameter(INT parameter, HVFCONTEXT context)
{
	VFCheckParameterType(parameter, VF_TYPE_WORD);
	WORD value;
	VFCheckResult(VFGetParameter(parameter, &value, context));
	return value;
}

SHORT VFGetShortParameter(INT parameter, HVFCONTEXT context)
{
	VFCheckParameterType(parameter, VF_TYPE_SHORT);
	SHORT value;
	VFCheckResult(VFGetParameter(parameter, &value, context));
	return value;
}

DWORD VFGetDWordParameter(INT parameter, HVFCONTEXT context)
{
	VFCheckParameterType(parameter, VF_TYPE_DWORD);
	DWORD value;
	VFCheckResult(VFGetParameter(parameter, &value, context));
	return value;
}

INT VFGetIntParameter(INT parameter, HVFCONTEXT context)
{
	VFCheckParameterType(parameter, VF_TYPE_INT);
	INT value;
	VFCheckResult(VFGetParameter(parameter, &value, context));
	return value;
}

bool VFGetBoolParameter(INT parameter, HVFCONTEXT context)
{
	VFCheckParameterType(parameter, VF_TYPE_BOOL);
	BOOL value;
	VFCheckResult(VFGetParameter(parameter, &value, context));
	return value != 0;
}

TCHAR VFGetCharParameter(INT parameter, HVFCONTEXT context)
{
	VFCheckParameterType(parameter, VF_TYPE_CHAR);
	CHAR value;
	VFCheckResult(VFGetParameter(parameter, &value, context));
	return (TCHAR)value;
}

string VFGetStringParameter(INT parameter, HVFCONTEXT context)
{
	VFCheckParameterType(parameter, VF_TYPE_STRING);
	INT l = VFGetParameter(parameter, NULL, context);
	VFCheckResult(l);
	CHAR *value;
	value = new CHAR[l + 1];
	l = VFGetParameter(parameter, value, context);
	string s;
	if(VFSucceeded(l)) s = AnsiToString(value);
	delete[] value;
	VFCheckResult(l);
	return s;
}

// Set

BYTE VFSetByteParameter(INT parameter, BYTE value, HVFCONTEXT context)
{
	BYTE oldValue = VFGetByteParameter(parameter, context);
	VFCheckResult(VFSetParameter(parameter, (DWORD)value, context));
	return oldValue;
}

SBYTE VFSetSByteParameter(INT parameter, SBYTE value, HVFCONTEXT context)
{
	SBYTE oldValue = VFGetSByteParameter(parameter, context);
	VFCheckResult(VFSetParameter(parameter, (DWORD)value, context));
	return oldValue;
}

WORD VFSetWordParameter(INT parameter, WORD value, HVFCONTEXT context)
{
	WORD oldValue = VFGetWordParameter(parameter, context);
	VFCheckResult(VFSetParameter(parameter, (DWORD)value, context));
	return oldValue;
}

SHORT VFSetShortParameter(INT parameter, SHORT value, HVFCONTEXT context)
{
	SHORT oldValue = VFGetShortParameter(parameter, context);
	VFCheckResult(VFSetParameter(parameter, (DWORD)value, context));
	return oldValue;
}

DWORD VFSetDWordParameter(INT parameter, DWORD value, HVFCONTEXT context)
{
	DWORD oldValue = VFGetDWordParameter(parameter, context);
	VFCheckResult(VFSetParameter(parameter, value, context));
	return oldValue;
}

INT VFSetIntParameter(INT parameter, INT value, HVFCONTEXT context)
{
	INT oldValue = VFGetIntParameter(parameter, context);
	VFCheckResult(VFSetParameter(parameter, (DWORD)value, context));
	return oldValue;
}

bool VFSetBoolParameter(INT parameter, bool value, HVFCONTEXT context)
{
	bool oldValue = VFGetBoolParameter(parameter, context);
	VFCheckResult(VFSetParameter(parameter, (DWORD)(BOOL)value, context));
	return oldValue;
}

TCHAR VFSetCharParameter(INT parameter, TCHAR value, HVFCONTEXT context)
{
	TCHAR oldValue = VFGetCharParameter(parameter, context);
	VFCheckResult(VFSetParameter(parameter, (DWORD)(CHAR)value, context));
	return oldValue;
}

string VFSetStringParameter(INT parameter, const string & value, HVFCONTEXT context)
{
	string oldValue = VFGetStringParameter(parameter, context);
#ifdef _UNICODE
	VFCheckResult(VFSetParameter(parameter, (DWORD)(LPCSTR)StringToAnsi(value), context));
#else
	VFCheckResult(VFSetParameter(parameter, (DWORD)(LPCSTR)value, context));
#endif
	return oldValue;
}

// Get general

INT VFGetParameterType(INT parameter)
{
	INT result = VFGetParameter(VFP_TYPE, (VOID*)parameter, NULL);
	VFCheckResult(result);
	return result;
}

string VFGetName()
{
	return VFGetStringParameter(VFP_NAME);
}

DWORD VFGetVersionHigh()
{
	return VFGetDWordParameter(VFP_VERSION_HIGH);
}

DWORD VFGetVersionLow()
{
	return VFGetDWordParameter(VFP_VERSION_LOW);
}

string VFGetCopyright()
{
	return VFGetStringParameter(VFP_COPYRIGHT);
}

//////////////////////////////////////////
//                                      //
// FAR and matching threshold functions //
//                                      //
//////////////////////////////////////////

const double ThFARLogRatio = 12;

double VFMatchingThresholdToFAR(INT th)
{
	if(th < 0) th = 0;
	return pow(10.0, -th / ThFARLogRatio + 2);
}

INT VFFARToMatchingThreshold(double f)
{
	if(f > 100.0) f = 100.0;
	else if(f <= 0.0) f = 1E-100;
	return Round((-log10(f) + 2) * ThFARLogRatio);
}
