using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.Text;
using System.Windows.Forms;
using Neurotec.Images;
using System.Diagnostics;

namespace Neurotec.Biometrics.VeriFingerDemo
{
	public partial class BatchEnrollForm : Form
	{
		#region Private fields

		private IList<string> fileNames;
		private IList<string> ids;
		private DateTime timeStarted;
		private bool finished;
		private int countEnrolled, countProcessed;
		private int aveSize;
		private TimeSpan aveTime;

		#endregion

		#region Public constructor

		public BatchEnrollForm()
		{
			InitializeComponent();
		}

		#endregion

		#region Private methods

		private void Log(string text)
		{
			richTextBox.AppendText(text);
			richTextBox.AppendText(Environment.NewLine);
			richTextBox.ScrollToCaret();
		}

		#endregion

		#region Public properties

		public IList<string> FileNames
		{
			get
			{
				return fileNames;
			}
			set
			{
				fileNames = value;
			}
		}

		public IList<string> Ids
		{
			get
			{
				return ids;
			}
			set
			{
				ids = value;
			}
		}

		#endregion

		private void backgroundWorker_DoWork(object sender, DoWorkEventArgs e)
		{
			int size = 0, c = 0;
			long time = 0;
			int count = fileNames.Count;
			int i;
			for (i = 0; i != count; i++)
			{
				if (backgroundWorker.CancellationPending) break;
				string fileName = fileNames[i];
				backgroundWorker.ReportProgress(i, string.Format("Loading image from file \"{0}\"", fileName));
				using (NImage image = NImage.FromFile(fileName))
				{
					float horzResolution = image.HorzResolution;
					float vertResolution = image.VertResolution;
					if (horzResolution < 250) horzResolution = 500;
					if (vertResolution < 250) vertResolution = 500;
					using (NGrayscaleImage grayImage = (NGrayscaleImage)NImage.FromImage(NPixelFormat.Grayscale, 0, horzResolution, vertResolution, image))
					{
						backgroundWorker.ReportProgress(i, string.Format("Image loaded. Size: {0}x{1}. Resolution: {2}x{3} dpi",
							grayImage.Width, grayImage.Height, grayImage.HorzResolution, grayImage.VertResolution));
						Stopwatch sw = Stopwatch.StartNew();
						byte[] template = Data.VFExtractor.Extract(grayImage, NFPosition.Unknown, NFImpressionType.LiveScanPlain);
						sw.Stop();
						time += sw.Elapsed.Ticks;
						if (template != null)
						{
							backgroundWorker.ReportProgress(i, string.Format("Template extracted. Size: {0}. Time: {1}",
								Data.SizeToString(template.Length), Data.TimeToString(sw.Elapsed)));
							c++;
							size += template.Length;
							Data.Database.Records.Add(ids[i], template);
						}
						else
						{
							backgroundWorker.ReportProgress(i, string.Format("Low quality image. Time: {0}",
								Data.TimeToString(sw.Elapsed)));
						}
					}
				}
			}
			countEnrolled = c;
			countProcessed = i;
			if (c != 0)
			{
				aveSize = (2 * size + c) / (2 * c);
				aveTime = TimeSpan.FromTicks((2 * time + i) / (2 * i));
			}
		}

		private void backgroundWorker_ProgressChanged(object sender, ProgressChangedEventArgs e)
		{
			Log((string)e.UserState);
			double p = (double)e.ProgressPercentage / fileNames.Count;
			progressBar.Value = (int)Math.Round(p * 100);
			TimeSpan timeElapsed = DateTime.Now - timeStarted;
			timeElapsedLabel.Text = Data.TimeToString(timeElapsed);
			if (p == 0)
			{
				timeRemainingLabel.Text = null;
			}
			else
			{
				TimeSpan timeRemaining = TimeSpan.FromTicks((long)Math.Round((double)timeElapsed.Ticks * (1 - p) / p));
				timeRemainingLabel.Text = Data.TimeToString(timeRemaining);
			}
		}

		private void backgroundWorker_RunWorkerCompleted(object sender, RunWorkerCompletedEventArgs e)
		{
			finished = true;
			closeButton.Text = "Close";
			closeButton.Enabled = true;
			timeElapsedLabel.Text = Data.TimeToString(DateTime.Now - timeStarted);
			timeRemainingLabel.Text = null;
			progressBar.Visible = false;
			if (e.Error != null)
			{
				string text = string.Format("Error: {0}", e.Error.Message);
				Log(text);
				MessageBox.Show(text, Text, MessageBoxButtons.OK, MessageBoxIcon.Error);
			}
			else
			{
				string text = string.Format("Enrollment {0}", countEnrolled == fileNames.Count ? "complete" : "stopped");
				if (countEnrolled == 0)
				{
					text += ". Nothing enrolled";
				}
				else
				{
					text += string.Format(". {0} of {1} template(s) enrolled. Average size: {2}. Average time: {3}",
						countEnrolled, countProcessed, Data.SizeToString(aveSize), Data.TimeToString(aveTime));
				}
				Log(text);
			}
		}

		private void BatchEnrollForm_FormClosing(object sender, FormClosingEventArgs e)
		{
			if (!finished)
			{
				closeButton.Enabled = false;
				backgroundWorker.CancelAsync();
				e.Cancel = true;
			}
		}

		private void BatchEnrollForm_Shown(object sender, EventArgs e)
		{
			finished = false;
			timeStarted = DateTime.Now;
			backgroundWorker.RunWorkerAsync();
		}
	}
}
