using System;
using System.Collections.Generic;
using System.Text;
using Neurotec.Biometrics.VeriFingerDemo.Properties;
using System.Reflection;
using System.Configuration;
using System.ComponentModel;
using System.Globalization;
using System.Collections.ObjectModel;
using System.IO;
using Neurotec.Images;

namespace Neurotec.Biometrics.VeriFingerDemo
{
	public class Record
	{
		#region Private fields

		private string id;
		byte[] template;

		#endregion

		#region Internal constructor

		internal Record(string id, byte[] template)
		{
			if (id == null) throw new ArgumentNullException("id");
			if (template == null) throw new ArgumentNullException("template");
			this.id = id;
			this.template = template;
		}

		#endregion

		#region Public properties

		public string ID
		{
			get
			{
				return id;
			}
		}

		public byte[] Template
		{
			get
			{
				return template;
			}
		}

		#endregion
	}

	public class Database : IDisposable
	{
		#region Public types

		public class RecordCollection: ReadOnlyCollection<Record>
		{
			#region Private fields

			private Database owner;
			private List<Record>[] itemsByG;

			#endregion

			#region Internal constructor

			internal RecordCollection(Database owner)
				: base(new List<Record>())
			{
				this.owner = owner;
				itemsByG = new List<Record>[256];
				for (int i = 0; i != 256; i++)
				{
					itemsByG[i] = new List<Record>();
				}
				for (; ; )
				{
					try
					{
						string id = owner.reader.ReadString();
						int templateLength = owner.reader.ReadInt32();
						byte[] template = owner.reader.ReadBytes(templateLength);
						int g = NFRecord.GetG(template);
						Record record = new Record(id, template);
						Items.Add(record);
						itemsByG[g].Add(record);
					}
					catch (EndOfStreamException)
					{
						break;
					}
				}
			}

			#endregion

			#region Public methods

			public Record Add(string id, byte[] template)
			{
				int g = NFRecord.GetG(template);
				Record record = new Record(id, template);
				owner.writer.Write(id);
				owner.writer.Write(template.Length);
				owner.writer.Write(template);
				owner.writer.Flush();
				Items.Add(record);
				itemsByG[g].Add(record);
				return record;
			}

			public void Clear()
			{
				owner.stream.SetLength(0);
				for (int i = 0; i != 256; i++)
				{
					itemsByG[i].Clear();
				}
				Items.Clear();
			}

			public IEnumerable<Record> GetEnumerator(byte[] template)
			{
				int g = NFRecord.GetG(template);
				int[] gs = new int[256];
				gs[0] = g;
				int count = 1;
				for (int i = 1; i != 256; i++)
				{
					int g1 = g + i;
					int g2 = g - i;
					if (g1 <= 255)
					{
						gs[count++] = g1;
					}
					if (g2 >= 0)
					{
						gs[count++] = g2;
					}
				}
				for (int i = 0; i != count; i++)
				{
					foreach (Record record in itemsByG[gs[i]])
					{
						yield return record;
					}
				}
			}

			#endregion

			#region Public properties

			public Record this[string id]
			{
				get
				{
					foreach(Record record in Items)
					{
						if (record.ID == id) return record;
					}
					return null;
				}
			}

			#endregion
		}

		#endregion

		#region Private fields

		private string fileName;
		private FileStream stream;
		private BinaryReader reader;
		private BinaryWriter writer;
		private RecordCollection records;

		#endregion

		#region Public constructor

		public Database(string fileName)
		{
			if (fileName == null) throw new ArgumentNullException("fileName");
			this.fileName = fileName;
			stream = new FileStream(fileName, FileMode.OpenOrCreate, FileAccess.ReadWrite);
			reader = new BinaryReader(stream);
			writer = new BinaryWriter(stream);
			records = new RecordCollection(this);
		}

		#endregion

		#region Destructor

		~Database()
		{
			Dispose();
		}

		#endregion

		#region Public properties

		public RecordCollection Records
		{
			get
			{
				return records;
			}
		}

		#endregion

		#region IDisposable Members

		public void Dispose()
		{
			if (stream != null)
			{
				reader.Close(); reader = null;
				writer.Close(); writer = null;
				stream.Close(); stream = null;
			}
		}

		#endregion
	}

	static class Data
	{
		#region Public static fields

		public static Database Database;
		public static VFExtractor VFExtractor;
		public static VFMatcher VFMatcher;

		#endregion

		#region Public static methods

		public static void ResetSetting(string name)
		{
			Settings settings = Settings.Default;
			PropertyInfo pi = typeof(Settings).GetProperty(name);
			DefaultSettingValueAttribute dsva = (DefaultSettingValueAttribute)pi.GetCustomAttributes(typeof(DefaultSettingValueAttribute), false)[0];
			pi.SetValue(settings, TypeDescriptor.GetConverter(pi.PropertyType).ConvertFromInvariantString(dsva.Value), null);
		}

		public static void UpdateVfe()
		{
			Settings settings = Settings.Default;
			try { VFExtractor.UseQuality = settings.VfeUseQuality; }
			catch { }
			try { VFExtractor.QualityThreshold = settings.VfeQualityThreshold; }
			catch { }
			try { VFExtractor.Mode = settings.VfeMode; }
			catch { }
			try { VFExtractor.TemplateSize = settings.VfeTemplateSize; }
			catch { }
			try { VFExtractor.ReturnedImage = settings.VfeReturnedImage; }
			catch { }
			try { VFExtractor.GeneralizationThreshold = settings.VfeGeneralizationThreshold; }
			catch { }
			try { VFExtractor.GeneralizationMaximalRotation = settings.VfeGeneralizationMaximalRotation; }
			catch { }
		}

		public static void UpdateVfeSettings()
		{
			Settings settings = Settings.Default;
			settings.VfeUseQuality = VFExtractor.UseQuality;
			settings.VfeQualityThreshold = VFExtractor.QualityThreshold;
			settings.VfeMode = VFExtractor.Mode;
			settings.VfeTemplateSize = VFExtractor.TemplateSize;
			settings.VfeReturnedImage = VFExtractor.ReturnedImage;
			settings.VfeGeneralizationThreshold = VFExtractor.GeneralizationThreshold;
			settings.VfeGeneralizationMaximalRotation = VFExtractor.GeneralizationMaximalRotation;
		}

		public static void UpdateVfm()
		{
			Settings settings = Settings.Default;
			try { VFMatcher.Mode = settings.VfmMode; }
			catch { }
			try { VFMatcher.MatchingThreshold = settings.VfmMatchingThreshold; }
			catch { }
			try { VFMatcher.MaximalRotation = settings.VfmMaximalRotation; }
			catch { }
		}

		public static void UpdateVfmSettings()
		{
			Settings settings = Settings.Default;
			settings.VfmMode = VFMatcher.Mode;
			settings.VfmMatchingThreshold = VFMatcher.MatchingThreshold;
			settings.VfmMaximalRotation = VFMatcher.MaximalRotation;
		}

		public static string MatchingThresholdToString(int value)
		{
			double p = -value / 12.0;
			return string.Format(string.Format("{{0:P{0}}}", Math.Max(0, (int)Math.Ceiling(-p) - 2)), Math.Pow(10, p));
		}

		public static int MatchingThresholdFromString(string value)
		{
			double p = Math.Log10(Math.Max(double.Epsilon, Math.Min(1,
				double.Parse(value.Replace(CultureInfo.CurrentCulture.NumberFormat.PercentSymbol, "")) / 100)));
			return Math.Max(0, (int)Math.Round(-12 * p));
		}

		public static int MaximalRotationToDegrees(byte value)
		{
			return (2 * value * 360 + 256) / (2 * 256);
		}

		public static byte MaximalRotationFromDegrees(int value)
		{
			return (byte)((2 * value * 256 + 360) / (2 * 360));
		}

		public static int QualityToPercent(byte value)
		{
			return (2 * value * 100 + 255) / (2 * 255);
		}

		public static byte QualityFromPercent(int value)
		{
			return (byte)((2 * value * 255 + 100) / (2 * 100));
		}

		public static string SizeToString(int size)
		{
			string value;
			int rem;
			size = Math.DivRem(size, 1024, out rem);
			value = rem + " byte(s)";
			if (size != 0)
			{
				size = Math.DivRem(size, 1024, out rem);
				value = rem + " KB " + value;
				if (size != 0)
				{
					size = Math.DivRem(size, 1024, out rem);
					value = rem + " MB " + value;
					if (size != 0)
					{
						size = Math.DivRem(size, 1024, out rem);
						value = rem + " GB " + value;
						if (size != 0)
						{
							value = size + " TB " + value;
						}
					}
				}
			}
			return value;
		}

		public static string TimeToString(TimeSpan time)
		{
			long t = time.Ticks / 10;
			string value;
			long rem;
			t = Math.DivRem(t, 1000, out rem);
			value = rem + " mks";
			if (t != 0)
			{
				t = Math.DivRem(t, 1000, out rem);
				value = rem + " ms " + value;
				if (t != 0)
				{
					t = Math.DivRem(t, 60, out rem);
					value = rem + " s " + value;
					if (t != 0)
					{
						t = Math.DivRem(t, 60, out rem);
						value = rem + " m " + value;
						if (t != 0)
						{
							t = Math.DivRem(t, 24, out rem);
							value = rem + " h " + value;
							if (t != 0)
							{
								value = t + " d " + value;
							}
						}
					}
				}
			}
			return value;
		}

		public static void GetImageFileFilters(out string openFilter, out string saveFilter)
		{
			StringBuilder openFileFilterBuilder = new StringBuilder();
			StringBuilder openAllFormatsFilterBuilder = new StringBuilder();
			StringBuilder saveFileFilterBuilder = new StringBuilder();
			foreach (NImageFormat imageFormat in NImageFormat.Formats)
			{
				if (imageFormat.CanRead)
				{
					openFileFilterBuilder.Append('|');
					openFileFilterBuilder.AppendFormat("{0} files ({1})|{1}",
						imageFormat.Name, imageFormat.FileFilter);
					if (openAllFormatsFilterBuilder.Length != 0) openAllFormatsFilterBuilder.Append(';');
					openAllFormatsFilterBuilder.Append(imageFormat.FileFilter);
				}
				if (imageFormat.CanWrite)
				{
					if (saveFileFilterBuilder.Length != 0) saveFileFilterBuilder.Append('|');
					saveFileFilterBuilder.AppendFormat("{0} files ({1})|{1}",
						imageFormat.Name, imageFormat.FileFilter);
				}
			}
			openFileFilterBuilder.Insert(0,
				string.Format("All supported images ({0})|{0}", openAllFormatsFilterBuilder.ToString()));
			openFileFilterBuilder.Append("|All files (*.*)|*.*");
			openFilter = openFileFilterBuilder.ToString();
			saveFilter = saveFileFilterBuilder.ToString();
		}

		public static string GetOpenImageFileFilter()
		{
			string openFilter, saveFilter;
			GetImageFileFilters(out openFilter, out saveFilter);
			return openFilter;
		}

		public static string GetSaveImageFileFilter()
		{
			string openFilter, saveFilter;
			GetImageFileFilters(out openFilter, out saveFilter);
			return saveFilter;
		}

		#endregion
	}

	public struct MatchingResult
	{
		#region Private fields

		private Record record;
		private int score;

		#endregion

		#region Public constructor

		public MatchingResult(Record record, int score)
		{
			if (record == null) throw new ArgumentNullException("record");
			this.record = record;
			this.score = score;
		}

		#endregion

		#region Public properties

		public Record Record
		{
			get
			{
				return record;
			}
		}

		public int Score
		{
			get
			{
				return score;
			}
		}

		public string RecordID
		{
			get
			{
				return record.ID;
			}
		}

		#endregion
	}
}
