using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.Text;
using System.Windows.Forms;
using System.Reflection;
using System.Runtime.InteropServices;
using System.Diagnostics;
using Neurotec.Biometrics.VeriFingerDemo.Properties;
using Neurotec.Images;
using Neurotec.Biometrics.Gui;
using System.IO;

namespace Neurotec.Biometrics.VeriFingerDemo
{
	public partial class MainForm : Form
	{
		#region External code

		[DllImport("kernel32.dll", CharSet = CharSet.Auto)]
		private static extern IntPtr LoadLibrary(string fileName);

		#endregion

		#region Private constants

		private const string appName = "VeriFinger Demo";

		#endregion

		#region Private fields

		private bool showComponents = false;
		private string status = null;
		private bool closing = false;
		private bool fpScannerManInitialized = false;
		private bool fpScannerManInitWork = false;
		private FPScannerMan fpScannerMan;

		#endregion

		#region Public constructor

		public MainForm()
		{
			InitializeComponent();

			UpdateShowComponents();

			componentsListView.BeginUpdate();
			AddComponentInfo(this.GetType());
			componentsListView.Items.Add(new ListViewItem(new string[]
				{
					"Microsoft .NET Framework",
					Environment.Version.ToString(),
					null,
					null
				}));
			componentsListView.Items.Add(new ListViewItem(new string[]
				{
					"    " + Environment.OSVersion.VersionString,
					Environment.OSVersion.Version.ToString(),
					null,
					null
				}));
			AddComponentInfo(typeof(NCore));
			AddComponentInfo(typeof(NImages));
			AddComponentInfo(typeof(FPScannerMan));
			AddComponentInfo(typeof(NFRecord));
			AddComponentInfo(typeof(NFView));
			AddComponentInfo(typeof(VFExtractor));
			AddComponentInfo(typeof(VFMatcher));
			componentColumnHeader.AutoResize(ColumnHeaderAutoResizeStyle.ColumnContent);
			versionColumnHeader.AutoResize(ColumnHeaderAutoResizeStyle.ColumnContent);
			copyrightColumnHeader.AutoResize(ColumnHeaderAutoResizeStyle.ColumnContent);
			componentsListView.EndUpdate();

			openFileDialog.Filter = Data.GetOpenImageFileFilter();

			Settings settings = Settings.Default;

			Data.VFExtractor = new VFExtractor();
			Data.UpdateVfe();
			Data.UpdateVfeSettings();

			Data.VFMatcher = new VFMatcher();
			Data.UpdateVfm();
			Data.UpdateVfmSettings();

			settings.Save();

			Data.Database = new Database(Path.ChangeExtension(Application.ExecutablePath, ".db"));
		}

		private void UpdateShowComponents()
		{
			componentsButton.Text = showComponents ? "&-" : "&+";
			refreshComponentsButton.Visible = componentsListView.Visible = showComponents;
			ClientSize = new Size(ClientSize.Width, (showComponents ? componentsListView.Bottom : componentsButton.Bottom) + 12);
		}

		#endregion

		#region Private methods

		private void RefreshComponentInfo(ListViewItem listViewItem)
		{
			PropertyInfo isRegisteredPropertyInfo = (PropertyInfo)listViewItem.Tag;
			if (isRegisteredPropertyInfo != null)
			{
				bool isRegistered = (bool)isRegisteredPropertyInfo.GetValue(null, null);
				listViewItem.SubItems[isRegisteredColumnHeader.Index].Text = isRegistered ? "Yes" : "No";
			}
		}

		private void AddComponentInfo(Type type)
		{
			PropertyInfo isRegisteredPropertyInfo = type.GetProperty("IsRegistered");
			if (isRegisteredPropertyInfo == null || isRegisteredPropertyInfo.PropertyType != typeof(bool) || !isRegisteredPropertyInfo.CanRead)
			{
				isRegisteredPropertyInfo = null;
			}

			string dllName = null;
			FieldInfo dllNameFieldInfo = type.GetField("DllName");
			if (dllNameFieldInfo != null && dllNameFieldInfo.FieldType == typeof(string) && dllNameFieldInfo.IsStatic)
			{
				dllName = (string)dllNameFieldInfo.GetValue(null);
			}

			Assembly assembly = type.Assembly;
			AssemblyTitleAttribute assemblyTitleAttribute = (AssemblyTitleAttribute)assembly.GetCustomAttributes(typeof(AssemblyTitleAttribute), false)[0];
			Version version = assembly.GetName().Version;
			AssemblyInformationalVersionAttribute[] assemblyInformationalVersionAttributes = (AssemblyInformationalVersionAttribute[])assembly.GetCustomAttributes(typeof(AssemblyInformationalVersionAttribute), false);
			AssemblyCopyrightAttribute assemblyCopyrightAttribute = (AssemblyCopyrightAttribute)assembly.GetCustomAttributes(typeof(AssemblyCopyrightAttribute), false)[0];

			FileVersionInfo dllVersionInfo = null;
			if (dllName != null)
			{
				IntPtr hModule = LoadLibrary(dllName);
				dllVersionInfo = FileVersionInfo.GetVersionInfo(dllName);
			}

			ListViewItem listViewItem = new ListViewItem(new string[]
				{
					assemblyTitleAttribute.Title,
					assemblyInformationalVersionAttributes.Length == 0 ? version.ToString() : assemblyInformationalVersionAttributes[0].InformationalVersion,
					null,
					assemblyCopyrightAttribute.Copyright
				});
			listViewItem.Tag = isRegisteredPropertyInfo;
			RefreshComponentInfo(listViewItem);
			componentsListView.Items.Add(listViewItem);
			if (dllVersionInfo != null)
			{
				componentsListView.Items.Add(new ListViewItem(new string[]
				{
					"    " + dllVersionInfo.FileDescription,
					new Version(dllVersionInfo.FileMajorPart, dllVersionInfo.FileMinorPart, dllVersionInfo.FileBuildPart, dllVersionInfo.FilePrivatePart).ToString(),
					null,
					dllVersionInfo.LegalCopyright
				}));
			}
		}

		private void UpdateStatus()
		{
			enrollButton.Enabled = verifyButton.Enabled = identifyButton.Enabled
				= useFPScannerManCheckBox.Enabled = batchEnrollButton.Enabled = !closing && !fpScannerManInitWork;
			fpScannersOptionsButton.Enabled = !closing && !fpScannerManInitWork && fpScannerManInitialized;
			enrollmentOptionsButton.Enabled = matchingOptionsButton.Enabled =
				databaseButton.Enabled = calculateRocButton.Enabled = !closing;
			Text = status == null ? appName : string.Format("{0} ({1})", appName, status);
		}

		private void UpdateStatus(string status)
		{
			this.status = status;
			UpdateStatus();
		}

		private void RequestClose()
		{
			if (!closing)
			{
				closing = true;
				UpdateStatus();
			}
		}

		private void CheckCloseRequest()
		{
			if (closing)
			{
				Close();
			}
		}

		private void InitFPScannerMan()
		{
			UseWaitCursor = true;
			fpScannerManInitWork = true;
			UpdateStatus("Initializing fingerprint scanners");
			fpScannerManInitBackgroundWorker.RunWorkerAsync();
		}

		private void UninitFPScannerMan()
		{
			UseWaitCursor = true;
			fpScannerManInitWork = true;
			UpdateStatus("Uninitializing fingerprint scanners");
			fpScannerManUninitBackgroundWorker.RunWorkerAsync();
		}

		private MatchingResult? Identify(byte[] template, bool matchAll, bool showAllResults, out int matchedCount, out int countThreshold, out List<MatchingResult> results, out TimeSpan matchTime, out TimeSpan seTime)
		{
			Settings settings = Settings.Default;
			bool useG = settings.FPUseG;
			int recordCount = Data.Database.Records.Count;
			countThreshold = useG
				? (recordCount * settings.FPGPercent + 99) / 100
				: recordCount;
			IEnumerable<Record> recordEnumerable = useG
				? Data.Database.Records.GetEnumerator(template)
				: Data.Database.Records;
			results = new List<MatchingResult>();
			int bestScore = 0;
			MatchingResult? bestResult = null;
			matchedCount = 0;
			Stopwatch sw = new Stopwatch();
			Stopwatch sesw = Stopwatch.StartNew();
			try
			{
				Data.VFMatcher.IdentifyStart(template);
			}
			catch (Exception ex)
			{
				throw new ApplicationException(string.Format("Error starting identification: {0}", ex.Message));
			}
			sesw.Stop();
			foreach (Record record in recordEnumerable)
			{
				if (matchedCount >= countThreshold) break;
				int score;
				sw.Start();
				try
				{
					score = Data.VFMatcher.IdentifyNext(record.Template);
				}
				catch (Exception ex)
				{
					throw new ApplicationException(string.Format("Error identifying: {0}", ex.Message));
				}
				sw.Stop();
				matchedCount++;
				MatchingResult? result = null;
				if (score != 0 || showAllResults)
				{
					result = new MatchingResult(record, score);
					results.Add(result.Value);
				}
				if (score != 0)
				{
					if (bestScore < score)
					{
						bestScore = score;
						bestResult = result;
						if (!matchAll) break;
					}
				}
			}
			sesw.Start();
			try
			{
				Data.VFMatcher.IdentifyEnd();
			}
			catch (Exception ex)
			{
				throw new ApplicationException(string.Format("Error ending identification: {0}", ex.Message));
			}
			sesw.Stop();
			matchTime = sw.Elapsed;
			seTime = sesw.Elapsed;
			return bestResult;
		}

		private void CalculateRoc(MatchListType rocType)
		{
			RocCalculationForm form = new RocCalculationForm();
			form.RocType = rocType;
			form.ShowDialog();
		}

		#endregion

		private void MainForm_FormClosing(object sender, FormClosingEventArgs e)
		{
			if (fpScannerManInitWork)
			{
				e.Cancel = true;
				RequestClose();
			}
			else if (fpScannerManInitialized)
			{
				e.Cancel = true;
				RequestClose();
				UninitFPScannerMan();
			}
			else
			{
				Data.VFExtractor.Dispose();
				Data.VFMatcher.Dispose();
			}
		}

		private void fpScannerManInitBackgroundWorker_DoWork(object sender, DoWorkEventArgs e)
		{
			fpScannerMan = new FPScannerMan(this);
		}

		private void fpScannerManInitBackgroundWorker_RunWorkerCompleted(object sender, RunWorkerCompletedEventArgs e)
		{
			UseWaitCursor = false;
			if (e.Error != null)
			{
				MessageBox.Show(string.Format("Fingerprint scanners initialization failed: {0}", e.Error.Message), appName,
					MessageBoxButtons.OK, MessageBoxIcon.Error);
			}
			else
			{
				fpScannerManInitialized = true;
			}
			fpScannerManInitWork = false;
			UpdateStatus(null);
			CheckCloseRequest();
		}

		private void fpScannerManUninitBackgroundWorker_DoWork(object sender, DoWorkEventArgs e)
		{
			fpScannerMan.Dispose();
			fpScannerMan = null;
		}

		private void fpScannerManUninitBackgroundWorker_RunWorkerCompleted(object sender, RunWorkerCompletedEventArgs e)
		{
			UseWaitCursor = false;
			fpScannerManInitialized = false;
			fpScannerManInitWork = false;
			UpdateStatus(null);
			CheckCloseRequest();
		}

		private void enrollButton_Click(object sender, EventArgs e)
		{
			NFRecordForm form = new NFRecordForm();
			form.Text = "Enroll";
			if (form.ShowDialog() == DialogResult.OK)
			{
				Settings settings = Settings.Default;
				if (settings.SearchForDuplicates)
				{
					List<MatchingResult> results;
					TimeSpan matchTime, seTime;
					int matchedCount, countThreshold;
					MatchingResult? result = Identify(form.Template, false, false,
						out matchedCount, out countThreshold, out results, out matchTime, out seTime);
					if (result != null)
					{
						MessageBox.Show(string.Format("Duplicate found: \"{0}\"", result.Value.RecordID), Text, MessageBoxButtons.OK, MessageBoxIcon.Error);
						return;
					}
				}
				Data.Database.Records.Add(form.ID, form.Template);
			}
		}

		private void verifyButton_Click(object sender, EventArgs e)
		{
			NFRecordForm form = new NFRecordForm();
			form.Text = "Verify";
			form.EnrollMode = false;
			if (form.ShowDialog() == DialogResult.OK)
			{
				Record record = Data.Database.Records[form.ID];
				if (record == null)
				{
					MessageBox.Show(string.Format("No template is enrolled with id \"{0}\" into the database", form.ID),
						Text, MessageBoxButtons.OK, MessageBoxIcon.Stop);
					return;
				}
				Stopwatch sw = Stopwatch.StartNew();
				int score;
				try
				{
					score = Data.VFMatcher.Verify(form.Template, record.Template);
				}
				catch (Exception ex)
				{
					MessageBox.Show(string.Format("Error verifying templates: {0}", ex.Message),
						Text, MessageBoxButtons.OK, MessageBoxIcon.Stop);
					return;
				}
				sw.Stop();
				MessageBox.Show(string.Format("Verification {0}. Time: {1}",
					score == 0 ? "failed" : string.Format("succeeded. Score: {0}", score), Data.TimeToString(sw.Elapsed)),
					Text, MessageBoxButtons.OK, score == 0 ? MessageBoxIcon.Stop : MessageBoxIcon.Information);
			}
		}

		private void identifyButton_Click(object sender, EventArgs e)
		{
			NFRecordForm form = new NFRecordForm();
			form.Text = "Identify";
			form.EnrollMode = false;
			form.ShowID = false;
			if (form.ShowDialog() == DialogResult.OK)
			{
				Settings settings = Settings.Default;
				List<MatchingResult> results;
				TimeSpan matchTime, seTime;
				int recordCount = Data.Database.Records.Count;
				int matchedCount, countThreshold;
				bool useG = settings.FPUseG;
				MatchingResult? result;
				try
				{
					result = Identify(form.Template, settings.MatchAllDatabase, settings.ShowAllMatchingResults,
						 out matchedCount, out countThreshold, out results, out matchTime, out seTime);
				}
				catch (Exception ex)
				{
					MessageBox.Show(ex.Message, Text, MessageBoxButtons.OK, MessageBoxIcon.Error);
					return;
				}
				string text;
				if (matchedCount == 0)
				{
					text = "Not identified. Nothing matched";
				}
				else
				{
					text = result == null ? "Not identified" : string.Format("Identified: \"{0}\"", result.Value.RecordID);
					text += string.Format(". {0} record(s) matched. Speed: {1} record(s) per second{2}. Start/end time: {3}",
						matchedCount, Math.Round(matchedCount * 1000 / matchTime.TotalMilliseconds),
						useG && (result != null || countThreshold != recordCount) ? string.Format(". Effective speed: {0} record(s) per second", Math.Round(recordCount * 1000 / matchTime.TotalMilliseconds)) : null,
						Data.TimeToString(seTime));
					if (result != null)
					{
						text += "\r\nShow identification details?";
					}
				}
				if (MessageBox.Show(text, Text,
					result == null ? MessageBoxButtons.OK : MessageBoxButtons.YesNo,
					result == null ? MessageBoxIcon.Stop : MessageBoxIcon.Information) == DialogResult.Yes)
				{
					IdentificationDetailsForm idForm = new IdentificationDetailsForm();
					idForm.Results = results;
					idForm.ShowDialog();
				}
			}
		}

		private void enrollmentOptionsButton_Click(object sender, EventArgs e)
		{
			EnrollmentOptionsForm form = new EnrollmentOptionsForm();
			form.ShowDialog();
		}

		private void useFPScannerManCheckBox_Click(object sender, EventArgs e)
		{
			Settings settings = Settings.Default;
			settings.UseFPScannerMan = !settings.UseFPScannerMan;
			settings.Save();
			if (settings.UseFPScannerMan)
			{
				if (!fpScannerManInitialized) InitFPScannerMan();
			}
			else
			{
				if (fpScannerManInitialized) UninitFPScannerMan();
			}
		}

		private void fpScannersOptionsButton_Click(object sender, EventArgs e)
		{
			FPScannersOptionsForm form = new FPScannersOptionsForm();
			form.ShowDialog();
		}

		private void matchingOptionsButton_Click(object sender, EventArgs e)
		{
			MatchingOptionsForm form = new MatchingOptionsForm();
			form.ShowDialog();
		}

		private void batchEnrollButton_Click(object sender, EventArgs e)
		{
			openFileDialog.FileName = null;
			if (openFileDialog.ShowDialog() == DialogResult.OK)
			{
				IdsForm idsForm = new IdsForm();
				idsForm.FileNames = openFileDialog.FileNames;
				if (idsForm.ShowDialog() == DialogResult.OK)
				{
					BatchEnrollForm form = new BatchEnrollForm();
					form.FileNames = openFileDialog.FileNames;
					form.Ids = idsForm.Ids;
					form.ShowDialog();
				}
			}
		}

		private void databaseButton_Click(object sender, EventArgs e)
		{
			DatabaseForm form = new DatabaseForm();
			form.ShowDialog();
		}

		private void calculateRocButton_Click(object sender, EventArgs e)
		{
			calculateRocContextMenuStrip.Show(calculateRocButton, 0, calculateRocButton.Height);
		}

		private void calculateRocFullToolStripMenuItem_Click(object sender, EventArgs e)
		{
			CalculateRoc(MatchListType.Full);
		}

		private void calculateRocHalfToolStripMenuItem_Click(object sender, EventArgs e)
		{
			CalculateRoc(MatchListType.Half);
		}

		private void calculateRocMinimalToolStripMenuItem_Click(object sender, EventArgs e)
		{
			CalculateRoc(MatchListType.Minimal);
		}

		private void componentsButton_Click(object sender, EventArgs e)
		{
			showComponents = !showComponents;
			UpdateShowComponents();
		}

		private void refreshComponentsButton_Click(object sender, EventArgs e)
		{
			componentsListView.BeginUpdate();
			foreach (ListViewItem listViewItem in componentsListView.Items)
			{
				RefreshComponentInfo(listViewItem);
			}
			componentsListView.EndUpdate();
		}

		private void MainForm_FormClosed(object sender, FormClosedEventArgs e)
		{
			Data.Database.Dispose();
		}

		private void MainForm_Shown(object sender, EventArgs e)
		{
			if (Settings.Default.UseFPScannerMan) InitFPScannerMan();
			else UpdateStatus();
		}
	}
}
