using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.Text;
using System.Windows.Forms;
using Neurotec.Biometrics.Gui;
using Neurotec.Images;
using Neurotec.Biometrics.VeriFingerDemo.Properties;
using System.Diagnostics;
using System.IO;

namespace Neurotec.Biometrics.VeriFingerDemo
{
	public partial class MainForm : Form
	{
		#region Private static readonly fields

		private static readonly NRgb resultImageMinColor = new NRgb(0, 230, 0);
		private static readonly NRgb resultImageMaxColor = new NRgb(255, 255, 255);

		#endregion

		#region Private fields

		private bool enrollMode = true;
		private FPScannerMan scannerMan;
		private NFView nfView1, nfView2;
		private NGrayscaleImage[] images = new NGrayscaleImage[0];
		private NGrayscaleImage[] resultImages = new NGrayscaleImage[0];
		private Bitmap[] bitmaps = new Bitmap[0];
		private Bitmap[] resultBitmaps = new Bitmap[0];
		private byte[][] templates = new byte[0][];
		private int count = 0;
		private int availableCount = 0;
		private NGrayscaleImage image;
		private NGrayscaleImage resultImage;
		private byte[] template;

        private enum ProgramMode{Enroll = 1, Verify = 2, Identify = 3};
        private ProgramMode mode;

        private const string appName = "VeriFinger 5.0 Demo - ";
		#endregion

		#region Public constructor

		public MainForm()
		{
			InitializeComponent();
		}

		#endregion

		#region Private methods

		private void scannerMan_ScannerAdded(object sender, FPScannerManScannerEventArgs e)
		{
            LogLine("Scanner plugged: {0}", false, e.Scanner.ID);
			Settings settings = Settings.Default;
			bool useSelectedScanners = settings.UseSelectedFPScanners;
			string[] selectedScanners = useSelectedScanners ?
				(settings.SelectedFPScanners == null ? null : settings.SelectedFPScanners.Split(Path.PathSeparator))
				: null;
			if (!useSelectedScanners || Array.IndexOf<string>(selectedScanners, e.Scanner.ID) != -1)
			{
				OnNewScanner(e.Scanner);
			}
		}

		private void scannerMan_ScannerRemoved(object sender, FPScannerManScannerEventArgs e)
		{
            LogLine("Scanner unplugged: {0}", false, e.Scanner.ID);
		}

		private void OnNewScanner(FPScanner scanner)
		{
			scanner.IsCapturingChanged += new EventHandler(scanner_IsCapturingChanged);
			scanner.FingerPlaced += new EventHandler(scanner_FingerPlaced);
			scanner.FingerRemoved += new EventHandler(scanner_FingerRemoved);
			scanner.ImageScanned += new FPScannerImageScannedEventHandler(scanner_ImageScanned);
			try
			{
				scanner.StartCapturing();
			}
			catch (Exception e)
			{
				string text = string.Format("Error starting capturing on scanner {0}: {1}", scanner.ID, e.Message);
				LogLine(text, false);
				MessageBox.Show(text, Text, MessageBoxButtons.OK, MessageBoxIcon.Error);
			}
		}

		private void scanner_IsCapturingChanged(object sender, EventArgs e)
		{
			FPScanner scanner = (FPScanner)sender;
			if (scanner.IsCapturing)
			{
				LogLine("Scanner {0} started capturing", false, scanner.ID);
			}
			else
			{
				LogLine("Scanner {0} stopped capturing", false, scanner.ID);
			}
		}

		private void scanner_FingerPlaced(object sender, EventArgs e)
		{
			FPScanner scanner = (FPScanner)sender;
            LogLine("Finger placed onto scanner {0}", false, scanner.ID);
		}

		private void scanner_FingerRemoved(object sender, EventArgs e)
		{
			FPScanner scanner = (FPScanner)sender;
            LogLine("Finger removed from scanner {0}", false, scanner.ID);
		}

		private void scanner_ImageScanned(object sender, FPScannerImageScannedEventArgs ea)
		{
			FPScanner scanner = (FPScanner)sender;
			OnImage((NGrayscaleImage)ea.Image.Clone(), string.Format("Image scanned from scanner {0}", scanner.ID), false);
		}

		private void OnImageChanged()
		{
            menuSaveImage.Enabled = image != null;
		}

		private void OnResultImageChanged()
		{
            menuSaveResultImage.Enabled = image != null;
		}

		private void ClearView()
		{
			nfView1.Image = null;
			if (nfView2.Template != null)
			{
				nfView2.Template.Dispose();
				nfView2.Template = null;
			}
			nfView2.Image = null;
			nfView2.ResultImage = null;
			image = null;
			resultImage = null;
			OnImageChanged();
			OnResultImageChanged();
		}

		private void Clear()
		{
			ClearView();
			Settings settings = Settings.Default;
			count = enrollMode && settings.FPUseGeneralization ? settings.FPGeneralizationTemplates : 1;
			foreach (NGrayscaleImage image in images)
			{
				if (image != null) image.Dispose();
			}
			foreach (NGrayscaleImage resultImage in resultImages)
			{
				if (resultImage != null) resultImage.Dispose();
			}
			foreach (Bitmap bitmap in bitmaps)
			{
				if (bitmap != null) bitmap.Dispose();
			}
			foreach (Bitmap resultBitmap in resultBitmaps)
			{
				if (resultBitmap != null) resultBitmap.Dispose();
			}
			images = new NGrayscaleImage[count];
			resultImages = new NGrayscaleImage[count];
			bitmaps = new Bitmap[count];
			resultBitmaps = new Bitmap[count];
			templates = new byte[count][];
			availableCount = 0;
			template = null;
			LogWait();
		}

		private void OnEnrollModeChanged()
		{
			Clear();
		}

		private void OnImage(NGrayscaleImage image, string logText, bool fromFile)
		{
			if (availableCount == count)
			{
				Clear();
			}
			else
			{
				ClearView();
			}
			LogLine(logText, true);

            if (!fromFile)
                LogLine(logText, false);

			LogLine("Size: {0}x{1}. Resolution: {2}x{3} dpi", true, image.Width, image.Height, image.HorzResolution, image.VertResolution);
			if (images[availableCount] != null) images[availableCount].Dispose();
			images[availableCount] = image;
			this.image = image;
			if (bitmaps[availableCount] != null) bitmaps[availableCount].Dispose();
			bitmaps[availableCount] = image.ToBitmap();
			nfView1.Image = bitmaps[availableCount];
			nfView2.Image = bitmaps[availableCount];
			OnImageChanged();

			UseWaitCursor = true;
			NGrayscaleImage resultImage = (NGrayscaleImage)image.Clone();
			Stopwatch sw = Stopwatch.StartNew();
			byte[] template;
			try
			{
				template = Data.VFExtractor.Extract(resultImage, NFPosition.Unknown, NFImpressionType.LiveScanPlain);
			}
			catch(Exception e)
			{
				string text = string.Format("Extraction error: {0}", e.Message);
				LogLine(text, true);
				MessageBox.Show(text, Text, MessageBoxButtons.OK, MessageBoxIcon.Error);
				return;
			}
			sw.Stop();
			UseWaitCursor = false;
			NFRecord templ = template == null ? null : new NFRecord(template);
			if (resultImages[availableCount] != null) resultImages[availableCount].Dispose();
			resultImages[availableCount] = resultImage;
			this.resultImage = resultImage;
			using (NImage ri = NImages.GetGrayscaleColorWrapper(resultImage, resultImageMinColor, resultImageMaxColor))
			{
				if (resultBitmaps[availableCount] != null) resultBitmaps[availableCount].Dispose();
				resultBitmaps[availableCount] = ri.ToBitmap();
			}
			templates[availableCount] = template;
			nfView2.ResultImage = resultBitmaps[availableCount];
			if (nfView2.Template != null) nfView2.Template.Dispose();
			nfView2.Template = templ;
			OnResultImageChanged();

			if (template == null)
			{
				string text = "Fingerprint image is of low quality";
				LogLine(text, true);
				MessageBox.Show(text, Text, MessageBoxButtons.OK, MessageBoxIcon.Error);
				return;
			}
			Settings settings = Settings.Default;
			if (settings.FPUseMinimalMinutiaCount && templ.Minutiae.Count < settings.FPMinimalMinutiaCount)
			{
				string text = string.Format("Fingerprint minutia count ({0}) is less than minimal value {1}", templ.Minutiae.Count, settings.FPMinimalMinutiaCount);
				LogLine(text, true);
				MessageBox.Show(text, Text, MessageBoxButtons.OK, MessageBoxIcon.Error);
				return;
			}
			LogLine("Template extracted{0}. G: {1}. Size: {2}. Time: {3}", true,
				Data.VFExtractor.UseQuality ? string.Format(". Quality: {0:P0}", Data.QualityToPercent(templ.Quality) / 100.0) : null,
				templ.G, Data.SizeToString(template.Length), Data.TimeToString(sw.Elapsed));
			availableCount++;

			if (availableCount == count)
			{
                lblWaitingForImg.Text = "Waiting for image(s)";

				if (enrollMode && settings.FPUseGeneralization)
				{
					Stopwatch swg = Stopwatch.StartNew();
					int baseTemplateIndex;
					try
					{
						template = Data.VFExtractor.Generalize(templates, out baseTemplateIndex);
					}
					catch (Exception e)
					{
						string text = string.Format("Generalization error: {0}", e.Message);
						LogLine(text, true);
						MessageBox.Show(text, Text, MessageBoxButtons.OK, MessageBoxIcon.Error);
						return;
					}
					swg.Stop();
					if (template == null)
					{
						string text = "Generalization failed";
                        LogLine(text, true);
						MessageBox.Show(text, Text, MessageBoxButtons.OK, MessageBoxIcon.Error);
						return;
					}
					templ = new NFRecord(template);
					LogLine("Template generalized, based on image #{0}{1}. G: {2}. Size: {3}. Time {4}",
						true, baseTemplateIndex + 1,
						Data.VFExtractor.UseQuality ? string.Format(". Quality: {0:P0}", Data.QualityToPercent(templ.Quality) / 100.0) : null,
						templ.G, Data.SizeToString(template.Length), Data.TimeToString(swg.Elapsed));
					nfView1.Image = bitmaps[baseTemplateIndex];
					nfView2.Image = bitmaps[baseTemplateIndex];
					nfView2.ResultImage = resultBitmaps[baseTemplateIndex];
					nfView2.Template.Dispose();
					nfView2.Template = templ;
					this.image = images[baseTemplateIndex];
					this.resultImage = resultImages[baseTemplateIndex];
					OnImageChanged();
					OnResultImageChanged();
				}
				this.template = template;

                switch (mode)
                {
                    case ProgramMode.Enroll:
                        doEnroll();
                        break;
                    case ProgramMode.Identify:
                        doIdentify();
                        break;
                    case ProgramMode.Verify:
                        doVerify();
                        break;
                }
			}
			else
			{
				LogWait();
			}

            panelNFView1_SizeChanged(null, null);
            panelNFView2_SizeChanged(null, null);
		}

        private void doEnroll()
        {
			Settings settings = Settings.Default;

            if (settings.SearchForDuplicates)
            {
                List<MatchingResult> results;
                TimeSpan matchTime, seTime;
                int matchedCount, countThreshold;
                bool canceled;

                MatchingResult? result = Identify(Template, false, false,
                    out matchedCount, out countThreshold, out results,
                    out matchTime, out seTime, out canceled);

                if (canceled)
                {
                    LogLine("Enrollment canceled", true);
                    return;
                }

                if (result != null)
                {
                    MessageBox.Show(string.Format("Duplicate found, record ID: \"{0}\"", result.Value.RecordID), Text, MessageBoxButtons.OK, MessageBoxIcon.Error);
                    return;
                }
            }

            EnterIDForm frm = new EnterIDForm();

            frm.setText("Enter template ID for enrollment");
            frm.ID = "";

            if (frm.ShowDialog() != DialogResult.OK)
            {
                frm.Dispose();
                return;
            }

            Data.Database.Records.Add(frm.ID, Template);
            frm.Dispose();
        }

        private void doVerify()
        {
            EnterIDForm frm = new EnterIDForm();
            string ID;

            frm.setText("Enter template ID for verification");
            frm.ID = "";

            if (frm.ShowDialog() != DialogResult.OK)
            {
                frm.Dispose();
                return;
            }

            ID = frm.ID;
            frm.Dispose();

            Record record = Data.Database.Records[ID];
            if (record == null)
            {
                MessageBox.Show(string.Format("No template is enrolled with id \"{0}\" into the database", ID),
                    Text, MessageBoxButtons.OK, MessageBoxIcon.Stop);

                return;
            }

            Stopwatch sw = Stopwatch.StartNew();
            int score;

            try
            {
                score = Data.VFMatcher.Verify(Template, record.Template);
            }
            catch (Exception ex)
            {
                MessageBox.Show(string.Format("Error verifying templates: {0}", ex.Message),
                    Text, MessageBoxButtons.OK, MessageBoxIcon.Stop);
                return;
            }

            sw.Stop();

            string str = string.Format("Verification {0}. Time: {1}",
                score == 0 ? "failed" : string.Format("succeeded. Score: {0}", score), Data.TimeToString(sw.Elapsed));
                
            MessageBox.Show(str, Text, MessageBoxButtons.OK, MessageBoxIcon.Information);

            LogLine(str, true);
        }

        private void doIdentify()
        {
            Settings settings = Settings.Default;
            List<MatchingResult> results;
            TimeSpan matchTime, seTime;
            int recordCount = Data.Database.Records.Count;
            int matchedCount, countThreshold;
            bool useG = settings.FPUseG;
            bool canceled = false;

            MatchingResult? result;

            try
            {
                result = Identify(Template, settings.MatchAllDatabase, settings.ShowAllMatchingResults,
                     out matchedCount, out countThreshold, out results,
                     out matchTime, out seTime, out canceled);

                if (canceled)
                {
                    LogLine("Identification canceled", true);
                    return;
                }
            }
            catch (Exception ex)
            {
                MessageBox.Show(ex.Message, Text, MessageBoxButtons.OK, MessageBoxIcon.Error);
                return;
            }

            dgvFields.Rows.Clear();

            string text;
            if (matchedCount == 0)
            {
                text = "Not identified. Nothing matched";
            }
            else
            {
                text = result == null ? "Not identified" : string.Format("Identified: \"{0}\"", result.Value.RecordID);
                text += string.Format(". {0} record(s) matched. Speed: {1} record(s) per second{2}. Start/end time: {3}",
                    matchedCount, Math.Round(matchedCount * 1000 / matchTime.TotalMilliseconds),
                    useG && (result != null || countThreshold != recordCount) ? string.Format(". Effective speed: {0} record(s) per second", Math.Round(recordCount * 1000 / matchTime.TotalMilliseconds)) : null,
                    Data.TimeToString(seTime));
            }

            LogLine(text, true);

            if (results != null)
            {
                foreach (MatchingResult res in results)
                {
                    dgvFields.Rows.Add(res.Score.ToString(), res.RecordID.ToString());
                }
            }
        }

		private void Log(string text, bool scroll, bool mainLog)
		{
            if (mainLog)
            {
                rtbMain.AppendText(text);
                if (scroll) rtbMain.ScrollToCaret();
            }
            else
            {
                rtbScanner.AppendText(text);
                if (scroll) rtbScanner.ScrollToCaret();
            }
		}

		private void LogLine(bool mainLog)
		{
            Log(Environment.NewLine, true, mainLog);
		}

		private void Log(string text, bool mainLog)
		{
            Log(text, true, mainLog);
		}

		private void LogLine(string text, bool mainLog)
		{
            Log(text, false, mainLog);
            LogLine(mainLog);
		}

		private void Log(string format, bool mainLog, params object[] args)
		{
            Log(string.Format(format, args), mainLog);
		}

        private void LogLine(string format, bool mainLog, params object[] args)
		{
            LogLine(string.Format(format, args), mainLog);
		}

		private void LogWait()
		{
			if (count == 1)
			{
                lblWaitingForImg.Text = "Waiting for image";
                LogLine(lblWaitingForImg.Text, true);
            }
			else
			{
                lblWaitingForImg.Text = string.Format("Waiting for image {0} of {1}", availableCount + 1, count);
                LogLine(lblWaitingForImg.Text, true);
			}
		}

		private void SaveImage(NGrayscaleImage image)
		{
			saveFileDialog.FileName = null;
			if (saveFileDialog.ShowDialog() == DialogResult.OK)
			{
				NImageFormat imageFormat = NImageFormat.Formats[saveFileDialog.FilterIndex - 1];
				string fileName = saveFileDialog.FileName;
				if (Path.GetExtension(fileName) == null)
				{
					fileName = fileName + '.' + imageFormat.DefaultFileExtension;
				}
				try
				{
					imageFormat.SaveImage(image, fileName);
				}
				catch (Exception ex)
				{
					MessageBox.Show(string.Format("Error saving to file \"{0}\": {1}", fileName, ex.Message),
						Text, MessageBoxButtons.OK, MessageBoxIcon.Error);
				}
			}
		}

		private void SetShownImage(ShownImage value)
		{
			Settings.Default.FPShownImage = nfView2.ShownImage = value;
            menuShowImgNone.Checked = Settings.Default.FPShownImage == ShownImage.None;
            menuShowImgResult.Checked = Settings.Default.FPShownImage == ShownImage.Result;
            menuShowImgOriginal.Checked = Settings.Default.FPShownImage == ShownImage.Original;
		}

		#endregion

		#region Public properties

		public bool EnrollMode
		{
			get
			{
				return enrollMode;
			}
			set
			{
                enrollMode = value;
				OnEnrollModeChanged();
			}
		}

		public byte[] Template
		{
			get
			{
				return template;
			}
		}

		#endregion

        private void openFile(bool promptForResolution)
        {
            openFileDialog.FileName = null;
            if (openFileDialog.ShowDialog() == DialogResult.OK)
            {
                int formatIndex = openFileDialog.FilterIndex - 2;
                NImageFormat imageFormat =
                    formatIndex == -1 || formatIndex == NImageFormat.Formats.Count
                    ? null : NImageFormat.Formats[formatIndex];
                string fileName = openFileDialog.FileName;

                NImage image;
                try
                {
                    image = NImage.FromFile(fileName, imageFormat);
                }
                catch (Exception ex)
                {
                    MessageBox.Show(string.Format("Error opening file \"{0}\": {1}", fileName, ex.Message),
                        Text, MessageBoxButtons.OK, MessageBoxIcon.Error);
                    return;
                }

                Settings settings = Settings.Default;
                float horzResolution = image.HorzResolution;
                float vertResolution = image.VertResolution;
                if (!promptForResolution)
                {
                    if (horzResolution < 250) horzResolution = 500;
                    if (vertResolution < 250) vertResolution = 500;
                }
                else
                {
                    ImageResolutionForm form = new ImageResolutionForm();
                    form.HorzResolution = horzResolution;
                    form.VertResolution = vertResolution;
                    if (form.ShowDialog() != DialogResult.OK)
                    {
                        return;
                    }
                    horzResolution = form.HorzResolution;
                    vertResolution = form.VertResolution;
                }
                NGrayscaleImage grayImage = (NGrayscaleImage)NImage.FromImage(NPixelFormat.Grayscale, 0, horzResolution, vertResolution, image);
                OnImage(grayImage, string.Format("Image loaded from file \"{0}\"", fileName), true);
            }
        }

		private void clearToolStripButton_Click(object sender, EventArgs e)
		{
			Clear();
		}

		private void NFRecordForm_FormClosed(object sender, FormClosedEventArgs e)
		{
            setStatus("Closing");

			if (scannerMan != null)
			{
				foreach(FPScanner scanner in scannerMan.Scanners)
				{
					scanner.StopCapturing();
				}
				scannerMan.Dispose();
			}

            Data.VFExtractor.Dispose();
            Data.VFMatcher.Dispose();
            Data.Database.Dispose();

			Settings settings = Settings.Default;
			settings.NFRecordFromSize = WindowState != FormWindowState.Normal ? RestoreBounds.Size : Size;
			settings.NFRecordFormMaximized = WindowState == FormWindowState.Maximized;
			Settings.Default.Save();
		}

		private void saveToolStripSplitButton_ButtonClick(object sender, EventArgs e)
		{
			SaveImage(image);
		}

		private void saveImageToolStripMenuItem_Click(object sender, EventArgs e)
		{
			SaveImage(image);
		}

		private void saveResultImageToolStripMenuItem_Click(object sender, EventArgs e)
		{
			SaveImage(resultImage);
		}

		private void shownImageNoneToolStripMenuItem_Click(object sender, EventArgs e)
		{
			SetShownImage(ShownImage.None);
		}

		private void shownImageOriginalToolStripMenuItem_Click(object sender, EventArgs e)
		{
			SetShownImage(ShownImage.Original);
		}

		private void shownImageResultToolStripMenuItem_Click(object sender, EventArgs e)
		{
			SetShownImage(ShownImage.Result);
		}

		private void showMinutiaeToolStripMenuItem_Click(object sender, EventArgs e)
		{
			Settings.Default.FPShowMinutiae = nfView2.ShowMinutiae = !nfView2.ShowMinutiae;
		}

		private void NFRecordForm_Shown(object sender, EventArgs e)
		{
			if (Settings.Default.NFRecordFormMaximized) WindowState = FormWindowState.Maximized;
		}

        private void menuClearImage_Click(object sender, EventArgs e)
        {
            Clear();
        }

        private void menuExit_Click(object sender, EventArgs e)
        {
            Close();
        }

        private void menuFileOpen_Click(object sender, EventArgs e)
        {
            openFile(false);
        }

        private void menuOpenWithPrompt_Click(object sender, EventArgs e)
        {
            openFile(true);
        }

        private void menuSaveImage_Click(object sender, EventArgs e)
        {
            SaveImage(image);
        }

        private void menuSaveResultImage_Click(object sender, EventArgs e)
        {
            SaveImage(resultImage);
        }

        private void menuEnrollmentOptions_Click(object sender, EventArgs e)
        {
            EnrollmentOptionsForm form = new EnrollmentOptionsForm();
            form.ShowDialog();
        }

        private void menuScannerOptions_Click(object sender, EventArgs e)
        {
            FPScannersOptionsForm form = new FPScannersOptionsForm();
            form.ShowDialog();
        }

        private void menuMatchingOptions_Click(object sender, EventArgs e)
        {
            MatchingOptionsForm form = new MatchingOptionsForm();
            form.ShowDialog();
        }

        private void menuBatchEnroll_Click(object sender, EventArgs e)
        {
            openFileDialog.FileName = null;
            if (openFileDialog.ShowDialog() == DialogResult.OK)
            {
                IdsForm idsForm = new IdsForm();
                idsForm.FileNames = openFileDialog.FileNames;
                if (idsForm.ShowDialog() == DialogResult.OK)
                {
                    BatchEnrollForm form = new BatchEnrollForm();
                    form.FileNames = openFileDialog.FileNames;
                    form.Ids = idsForm.Ids;
                    form.ShowDialog();
                }
            }
        }

        private void menuViewDatabase_Click(object sender, EventArgs e)
        {
            DatabaseForm form = new DatabaseForm();
            form.ShowDialog();
        }

        private void CalculateRoc(MatchListType rocType)
        {
            RocCalculationForm form = new RocCalculationForm();
            form.RocType = rocType;
            form.ShowDialog();
        }

        private void menuROCMinimal_Click(object sender, EventArgs e)
        {
            CalculateRoc(MatchListType.Minimal);
        }

        private void menuROCHalf_Click(object sender, EventArgs e)
        {
            CalculateRoc(MatchListType.Half);
        }

        private void menuROCFull_Click(object sender, EventArgs e)
        {
            CalculateRoc(MatchListType.Full);
        }

        private void menuShowMinutae_Click(object sender, EventArgs e)
        {
            Settings.Default.FPShowMinutiae = nfView2.ShowMinutiae = !nfView2.ShowMinutiae;
            menuShowMinutae.Checked = Settings.Default.FPShowMinutiae;
        }

        private void menuShowImgNone_Click(object sender, EventArgs e)
        {
            SetShownImage(ShownImage.None);
        }

        private void menuShowImgOriginal_Click(object sender, EventArgs e)
        {
            SetShownImage(ShownImage.Original);
        }

        private void menuShowImgResult_Click(object sender, EventArgs e)
        {
            SetShownImage(ShownImage.Result);
        }

        private void panelNFView1_SizeChanged(object sender, EventArgs e)
        {
            if ((nfView1 != null) && (nfView1.Image != null))
            {
                nfView1.Width = nfView1.Image.Width;
                nfView1.Height = nfView1.Image.Height;
                nfView1.Left = (nfViewSplitContainer.Panel1.Width - nfView1.Width) / 2;
                nfView1.Top = (nfViewSplitContainer.Panel1.Height - nfView1.Height) / 2;
            }
        }

        private void panelNFView2_SizeChanged(object sender, EventArgs e)
        {
            if ((nfView2 != null) && (nfView2.Image != null))
            {
                nfView2.Width = nfView2.Image.Width;
                nfView2.Height = nfView2.Image.Height;
                nfView2.Left = (nfViewSplitContainer.Panel2.Width - nfView2.Width) / 2;
                nfView2.Top = (nfViewSplitContainer.Panel2.Height - nfView2.Height) / 2;
            }
        }

        private void modeCheck(MenuItem itemToCheck)
        {
            menuModeEnroll.Checked =
                menuModeVerify.Checked =
                menuModeIdentify.Checked = false;

            itemToCheck.Checked = true;
        }

        private void menuModeEnroll_Click(object sender, EventArgs e)
        {
            Settings.Default.ProgramMode = (int)ProgramMode.Enroll;
            modeCheck(menuModeEnroll);
            setMode(ProgramMode.Enroll);
        }

        private void menuModeVerify_Click(object sender, EventArgs e)
        {
            Settings.Default.ProgramMode = (int)ProgramMode.Verify;
            modeCheck(menuModeVerify);
            setMode(ProgramMode.Verify);
        }

        private void menuModeIdentify_Click(object sender, EventArgs e)
        {
            Settings.Default.ProgramMode = (int)ProgramMode.Identify;
            modeCheck(menuModeIdentify);
            setMode(ProgramMode.Identify);
        }

        private void NFRecordForm_Load(object sender, EventArgs e)
        {
            setStatus("Initializing...");

            Settings settings = Settings.Default;

            Data.VFExtractor = new VFExtractor();
            Data.UpdateVfe();
            Data.UpdateVfeSettings();

            Data.VFMatcher = new VFMatcher();
            Data.UpdateVfm();
            Data.UpdateVfmSettings();

            settings.Save();

            Data.Database = new Database(Path.ChangeExtension(Application.ExecutablePath, ".db"));

            Size = settings.NFRecordFromSize;
            mainSplitContainer.Height = this.ClientSize.Height - statusStrip.Height - 3;

            nfView1 = new NFView();
            nfView1.BorderStyle = BorderStyle.None;
            panelNFView1.Controls.Add(nfView1);
            nfView1.Width = nfView1.Parent.Width;
            nfView1.Height = nfView1.Parent.Height;

            nfView2 = new NFView();
            nfView2.BorderStyle = BorderStyle.None;
            nfView2.ShownImage = ShownImage.Result;
            nfView2.AllowSelection = false;
            nfView2.AllowHover = false;
            panelNFView2.Controls.Add(nfView2);
            nfView2.ShownImage = settings.FPShownImage;
            nfView2.ShowMinutiae = settings.FPShowMinutiae;
            nfView2.Width = nfView2.Parent.Width;
            nfView2.Height = nfView2.Parent.Height;

            string openFilter, saveFilter;
            Data.GetImageFileFilters(out openFilter, out saveFilter);
            openFileDialog.Filter = openFilter;
            saveFileDialog.Filter = saveFilter;

            if (settings.UseFPScannerMan)
            {
                bool useSelectedScanners = settings.UseSelectedFPScanners;
                string[] selectedScanners = useSelectedScanners ?
                    (settings.SelectedFPScanners == null ? null : settings.SelectedFPScanners.Split(Path.PathSeparator))
                    : null;

                InitialisingForm frm = new InitialisingForm();
                frm.Show();
                frm.Refresh();

                try
                {
                    scannerMan = new FPScannerMan(this);
                }
                catch (Exception ex)
                {
                    throw ex;
                }
                finally
                {
                    frm.Close();
                    frm.Dispose();
                }

                scannerMan.ScannerAdded += new FPScannerManScannerEventHandler(scannerMan_ScannerAdded);
                scannerMan.ScannerRemoved += new FPScannerManScannerEventHandler(scannerMan_ScannerRemoved);
                foreach (FPScanner scanner in scannerMan.Scanners)
                {
                    if (!useSelectedScanners || Array.IndexOf<string>(selectedScanners, scanner.ID) != -1)
                    {
                        OnNewScanner(scanner);
                    }
                }
            }

            ProgramMode mode = (ProgramMode)settings.ProgramMode;
            switch (mode)
            {
                case ProgramMode.Enroll:
                    menuModeEnroll_Click(null, null);
                    break;
                case ProgramMode.Identify:
                    menuModeIdentify_Click(null, null);
                    break;
                case ProgramMode.Verify:
                    menuModeVerify_Click(null, null);
                    break;
            }
        }

        private void setStatus(string status)
        {
            Text = appName + status;
        }

        private void setMode(ProgramMode mode)
        {
            setStatus(mode.ToString());

            switch (mode)
            {
                case ProgramMode.Enroll:
                    EnrollMode = true;
                    break;
                case ProgramMode.Identify:
                    EnrollMode = false;
                    break;
                case ProgramMode.Verify:
                    EnrollMode = false;
                    break;
            }

            this.mode = mode;
            dgvFields.Rows.Clear();
        }

        private MatchingResult? Identify(byte[] template, bool matchAll, bool showAllResults, out int matchedCount, out int countThreshold, out List<MatchingResult> results, out TimeSpan matchTime, out TimeSpan seTime, out bool canceled)
        {
            Settings settings = Settings.Default;

            ProgressForm frm = new ProgressForm();
            frm.setTitle(Text);
            frm.setProgress(0);

            frm.Show(this);
            frm.Refresh();

            canceled = false;

            bool useG = settings.FPUseG;
            int recordCount = Data.Database.Records.Count;

            countThreshold = useG
                ? (recordCount * settings.FPGPercent + 99) / 100
                : recordCount;

            IEnumerable<Record> recordEnumerable = useG
                ? Data.Database.Records.GetEnumerator(template)
                : Data.Database.Records;

            results = new List<MatchingResult>();

            int bestScore = 0;

            MatchingResult? bestResult = null;
            matchedCount = 0;
            Stopwatch sw = new Stopwatch();
            Stopwatch sesw = Stopwatch.StartNew();

            try
            {
                Data.VFMatcher.IdentifyStart(template);
            }
            catch (Exception ex)
            {
                frm.Close();
                frm.Dispose();
                throw new ApplicationException(string.Format("Error starting identification: {0}", ex.Message));
            }

            sesw.Stop();

            foreach (Record record in recordEnumerable)
            {
                frm.setProgress(100 * matchedCount / recordCount);
                frm.Refresh();
                Application.DoEvents();

                if (frm.getStopPressed())
                {
                    frm.Close();
                    frm.Dispose();
                    canceled = true;
                    results.Clear();
                    matchTime = System.TimeSpan.Zero;
                    seTime = matchTime;

                    try
                    {
                        Data.VFMatcher.IdentifyEnd();
                    }
                    catch (Exception ex)
                    {
                        throw new ApplicationException(string.Format("Error ending identification: {0}", ex.Message));
                    }

                    return null;
                }

                if (matchedCount >= countThreshold) break;
                int score;
                sw.Start();
                try
                {
                    score = Data.VFMatcher.IdentifyNext(record.Template);
                }
                catch (Exception ex)
                {
                    frm.Close();
                    frm.Dispose();
                    throw new ApplicationException(string.Format("Error identifying: {0}", ex.Message));
                }
                sw.Stop();
                matchedCount++;
                MatchingResult? result = null;
                if (score != 0 || showAllResults)
                {
                    result = new MatchingResult(record, score);
                    results.Add(result.Value);
                }
                if (score != 0)
                {
                    if (bestScore < score)
                    {
                        bestScore = score;
                        bestResult = result;
                        if (!matchAll) break;
                    }
                }
            }

            frm.Close();
            frm.Dispose();

            sesw.Start();
            try
            {
                Data.VFMatcher.IdentifyEnd();
            }
            catch (Exception ex)
            {
                throw new ApplicationException(string.Format("Error ending identification: {0}", ex.Message));
            }
            
            sesw.Stop();

            matchTime = sw.Elapsed;
            seTime = sesw.Elapsed;

            return bestResult;
        }

        private void menuItem6_Click(object sender, EventArgs e)
        {
            AboutForm frm = new AboutForm();
            frm.ShowDialog();
            frm.Dispose();
        }
	}
}
